import os

from collections.abc import Callable

import pytest


@pytest.fixture
def example() -> Callable[[str], str]:
    def _example(name: str) -> str:
        with open(
            os.path.join(os.path.dirname(__file__), "examples", name + ".toml"),
            encoding="utf-8",
        ) as f:
            return f.read()

    return _example


@pytest.fixture
def json_example() -> Callable[[str], str]:
    def _example(name: str) -> str:
        with open(
            os.path.join(os.path.dirname(__file__), "examples", "json", name + ".json"),
            encoding="utf-8",
        ) as f:
            return f.read()

    return _example


@pytest.fixture
def invalid_example() -> Callable[[str], str]:
    def _example(name: str) -> str:
        with open(
            os.path.join(
                os.path.dirname(__file__), "examples", "invalid", name + ".toml"
            ),
            encoding="utf-8",
        ) as f:
            return f.read()

    return _example



from datetime import date
from datetime import datetime as dt
from datetime import time
from datetime import timedelta as td
from datetime import timezone as tz
from typing import Union

import pytest

from tomlkit._utils import _utc
from tomlkit._utils import parse_rfc3339


@pytest.mark.parametrize(
    "string, expected",
    [
        ("1979-05-27T07:32:00", dt(1979, 5, 27, 7, 32, 0)),
        ("1979-05-27T07:32:00Z", dt(1979, 5, 27, 7, 32, 0, tzinfo=_utc)),
        (
            "1979-05-27T07:32:00-07:00",
            dt(1979, 5, 27, 7, 32, 0, tzinfo=tz(td(seconds=-7 * 3600), "-07:00")),
        ),
        (
            "1979-05-27T00:32:00.999999-07:00",
            dt(
                1979,
                5,
                27,
                0,
                32,
                0,
                999999,
                tzinfo=tz(td(seconds=-7 * 3600), "-07:00"),
            ),
        ),
    ],
)
def test_parse_rfc3339_datetime(string: str, expected: Union[dt, date, time]) -> None:
    assert parse_rfc3339(string) == expected


@pytest.mark.parametrize("string, expected", [("1979-05-27", date(1979, 5, 27))])
def test_parse_rfc3339_date(string: str, expected: Union[dt, date, time]) -> None:
    assert parse_rfc3339(string) == expected


@pytest.mark.parametrize(
    "string, expected",
    [("12:34:56", time(12, 34, 56)), ("12:34:56.123456", time(12, 34, 56, 123456))],
)
def test_parse_rfc3339_time(string: str, expected: Union[dt, date, time]) -> None:
    assert parse_rfc3339(string) == expected



from __future__ import annotations

import copy
import ctypes
import math
import pickle
import sys

from collections.abc import Callable
from datetime import date
from datetime import datetime
from datetime import time
from datetime import timedelta
from datetime import timezone
from datetime import tzinfo
from typing import Any

import pytest

from tests.util import assert_is_ppo
from tests.util import elementary_test
from tomlkit import api
from tomlkit import parse
from tomlkit.container import OutOfOrderTableProxy
from tomlkit.exceptions import NonExistentKey
from tomlkit.items import Array
from tomlkit.items import Bool
from tomlkit.items import Comment
from tomlkit.items import InlineTable
from tomlkit.items import Integer
from tomlkit.items import Item
from tomlkit.items import KeyType
from tomlkit.items import Null
from tomlkit.items import SingleKey as Key
from tomlkit.items import String
from tomlkit.items import StringType
from tomlkit.items import Table
from tomlkit.items import Trivia
from tomlkit.items import item
from tomlkit.parser import Parser


@pytest.fixture()
def tz_pst() -> tzinfo:
    try:
        from datetime import timezone

        return timezone(timedelta(hours=-8), "PST")
    except ImportError:
        from datetime import tzinfo as _tzinfo

        class PST(_tzinfo):
            def utcoffset(self, dt: datetime | None) -> timedelta:
                return timedelta(hours=-8)

            def tzname(self, dt: datetime | None) -> str:
                return "PST"

            def dst(self, dt: datetime | None) -> timedelta:
                return timedelta(0)

        return PST()


@pytest.fixture()
def tz_utc() -> tzinfo:
    try:
        from datetime import timezone

        return timezone.utc
    except ImportError:
        from datetime import tzinfo as _tzinfo

        class UTC(_tzinfo):
            def utcoffset(self, dt: datetime | None) -> timedelta:
                return timedelta(hours=0)

            def tzname(self, dt: datetime | None) -> str:
                return "UTC"

            def dst(self, dt: datetime | None) -> timedelta:
                return timedelta(0)

        return UTC()


def test_item_base_has_no_unwrap() -> None:
    trivia = Trivia(indent="\t", comment_ws=" ", comment="For unit test")
    item = Item(trivia)
    try:
        item.unwrap()
    except NotImplementedError:
        pass
    else:
        raise AssertionError("`items.Item` should not implement `unwrap`")


def test_integer_unwrap() -> None:
    elementary_test(item(666), int)


def test_float_unwrap() -> None:
    elementary_test(item(2.78), float)


@pytest.mark.skipif(
    sys.implementation.name != "cpython", reason="PySequence_Check is CPython-specific"
)
def test_float_is_not_a_sequence() -> None:
    value = parse("a = [1.0, 2.0, 3.0]")["a"][0]
    py_sequence_check = ctypes.pythonapi.PySequence_Check
    py_sequence_check.argtypes = [ctypes.py_object]
    py_sequence_check.restype = ctypes.c_int

    assert not py_sequence_check(value)


def test_false_unwrap() -> None:
    elementary_test(item(False), bool)


def test_true_unwrap() -> None:
    elementary_test(item(True), bool)


def test_datetime_unwrap() -> None:
    dt = datetime.now(tz=timezone.utc)
    elementary_test(item(dt), datetime)


def test_string_unwrap() -> None:
    elementary_test(item("hello"), str)


def test_null_unwrap() -> None:
    n = Null()
    elementary_test(n, type(None))


def test_aot_unwrap() -> None:
    d = item([{"a": "A"}, {"b": "B"}])
    unwrapped = d.unwrap()
    assert_is_ppo(unwrapped, list)
    for du, _ in zip(unwrapped, d):  # noqa: B905
        assert_is_ppo(du, dict)
        for ku in du:
            vu = du[ku]
            assert_is_ppo(ku, str)
            assert_is_ppo(vu, str)


def test_aot_set_item() -> None:
    d = item(["A", {"b": "B"}, ["c", "D"]])
    d[0] = "C"
    assert isinstance(d[0], String)
    assert d[0] == "C"
    d[1]["b"] = "D"
    assert isinstance(d[1], InlineTable)
    assert d[1]["b"] == "D"
    d[0] = ["c", "C"]
    assert isinstance(d[0], Array)
    assert d[0][1] == "C"


def test_time_unwrap() -> None:
    t = time(3, 8, 14)
    elementary_test(item(t), time)


def test_date_unwrap() -> None:
    d = date.today()
    elementary_test(item(d), date)


def test_array_unwrap() -> None:
    trivia = Trivia(indent="\t", comment_ws=" ", comment="For unit test")
    i = item(666)
    f = item(2.78)
    b = item(False)
    a = Array([i, f, b], trivia)
    a_unwrapped = a.unwrap()
    assert_is_ppo(a_unwrapped, list)
    assert_is_ppo(a_unwrapped[0], int)
    assert_is_ppo(a_unwrapped[1], float)
    assert_is_ppo(a_unwrapped[2], bool)


def test_abstract_table_unwrap() -> None:
    table = item({"foo": "bar"})
    super_table = item({"table": table, "baz": "borg"})

    table_unwrapped = super_table.unwrap()
    sub_table = table_unwrapped["table"]
    assert_is_ppo(table_unwrapped, dict)
    assert_is_ppo(sub_table, dict)
    for ku in sub_table:
        vu = sub_table[ku]
        assert_is_ppo(ku, str)
        assert_is_ppo(vu, str)


def test_key_comparison() -> None:
    k = Key("foo")

    assert k == Key("foo")
    assert k == "foo"
    assert k != "bar"
    assert k != 5


def test_items_can_be_appended_to_and_removed_from_a_table() -> None:
    string = """[table]
"""

    parser = Parser(string)
    _, table = parser._parse_table()

    assert isinstance(table, Table)
    assert table.as_string() == ""

    table.append(Key("foo"), String(StringType.SLB, "bar", "bar", Trivia(trail="\n")))

    assert table.as_string() == 'foo = "bar"\n'

    table.append(
        Key("baz"),
        Integer(34, Trivia(comment_ws="   ", comment="# Integer", trail=""), "34"),
    )

    assert table.as_string() == 'foo = "bar"\nbaz = 34   # Integer'

    table.remove(Key("baz"))

    assert table.as_string() == 'foo = "bar"\n'

    table.remove(Key("foo"))

    assert table.as_string() == ""

    with pytest.raises(NonExistentKey):
        table.remove(Key("foo"))


def test_items_can_be_appended_to_and_removed_from_an_inline_table() -> None:
    string = """table = {}
"""

    parser = Parser(string)
    result = parser._parse_item()
    assert result is not None
    _, table = result

    assert isinstance(table, InlineTable)
    assert table.as_string() == "{}"

    table.append(Key("foo"), String(StringType.SLB, "bar", "bar", Trivia(trail="")))

    assert table.as_string() == '{foo = "bar"}'

    table.append(Key("baz"), Integer(34, Trivia(trail=""), "34"))

    assert table.as_string() == '{foo = "bar", baz = 34}'

    table.remove(Key("baz"))

    assert table.as_string() == '{foo = "bar"}'

    table.remove(Key("foo"))

    assert table.as_string() == "{}"

    with pytest.raises(NonExistentKey):
        table.remove(Key("foo"))


def test_inf_and_nan_are_supported(example: Callable[[str], str]) -> None:
    content = example("0.5.0")
    doc = parse(content)

    assert doc["sf1"] == float("inf")
    assert doc["sf2"] == float("inf")
    assert doc["sf3"] == float("-inf")

    assert math.isnan(doc["sf4"])
    assert math.isnan(doc["sf5"])
    assert math.isnan(doc["sf6"])


def test_hex_octal_and_bin_integers_are_supported(
    example: Callable[[str], str],
) -> None:
    content = example("0.5.0")
    doc = parse(content)

    assert doc["hex1"] == 3735928559
    assert doc["hex2"] == 3735928559
    assert doc["hex3"] == 3735928559

    assert doc["oct1"] == 342391
    assert doc["oct2"] == 493

    assert doc["bin1"] == 214


def test_key_automatically_sets_proper_string_type_if_not_bare() -> None:
    key = Key("foo.bar")

    assert key.t == KeyType.Basic

    key = Key("")
    assert key.t == KeyType.Basic


def test_array_behaves_like_a_list() -> None:
    a = item([1, 2])

    assert a == [1, 2]
    assert a.as_string() == "[1, 2]"

    a += [3, 4]
    assert a == [1, 2, 3, 4]
    assert a.as_string() == "[1, 2, 3, 4]"

    del a[2]
    assert a == [1, 2, 4]
    assert a.as_string() == "[1, 2, 4]"

    assert a.pop() == 4
    assert a == [1, 2]
    assert a.as_string() == "[1, 2]"

    a[0] = 4
    assert a == [4, 2]
    a[-2] = 0
    assert a == [0, 2]

    del a[-2]
    assert a == [2]
    assert a.as_string() == "[2]"

    a.clear()
    assert a == []
    assert a.as_string() == "[]"

    content = """a = [1, 2,] # Comment
"""
    doc = parse(content)
    assert str(doc["a"]) == "[1, 2]"

    assert doc["a"] == [1, 2]
    doc["a"] += [3, 4]
    assert doc["a"] == [1, 2, 3, 4]
    assert (
        doc.as_string()
        == """a = [1, 2, 3, 4] # Comment
"""
    )


def test_array_multiline() -> None:
    t = item([1, 2, 3, 4, 5, 6, 7, 8])
    t.multiline(True)

    expected = """\
[
    1,
    2,
    3,
    4,
    5,
    6,
    7,
    8,
]"""

    assert expected == t.as_string()

    t2: Any = item([])
    t2.multiline(True)

    assert t2.as_string() == "[]"


def test_array_multiline_modify() -> None:
    doc = parse(
        """\
a = [
    "abc"
]"""
    )
    doc["a"].append("def")
    expected = """\
a = [
    "abc",
    "def"
]"""
    assert expected == doc.as_string()
    doc["a"].insert(1, "ghi")
    expected = """\
a = [
    "abc",
    "ghi",
    "def"
]"""
    assert expected == doc.as_string()


def test_append_to_empty_array() -> None:
    doc = parse("x = [ ]")
    doc["x"].append("a")
    assert doc.as_string() == 'x = ["a" ]'
    doc = parse("x = [\n]")
    doc["x"].append("a")
    assert doc.as_string() == 'x = [\n    "a"\n]'


def test_modify_array_with_comment() -> None:
    doc = parse("x = [ # comment\n]")
    doc["x"].append("a")
    assert doc.as_string() == 'x = [ # comment\n    "a"\n]'
    doc = parse(
        """\
x = [
    "a",
    # comment
    "b"
]"""
    )
    doc["x"].insert(1, "c")
    expected = """\
x = [
    "a",
    # comment
    "c",
    "b"
]"""
    assert doc.as_string() == expected
    doc = parse(
        """\
x = [
    1  # comment
]"""
    )
    doc["x"].append(2)
    assert (
        doc.as_string()
        == """\
x = [
    1,  # comment
    2
]"""
    )
    doc["x"].pop(0)
    assert doc.as_string() == "x = [\n    2\n]"


def test_append_to_multiline_array_with_comment() -> None:
    doc = parse(
        """\
x = [
    # Here is a comment
    1,
    2
]
"""
    )
    doc["x"].multiline(True).append(3)
    assert (
        doc.as_string()
        == """\
x = [
    # Here is a comment
    1,
    2,
    3,
]
"""
    )
    assert doc["x"].pop() == 3
    assert (
        doc.as_string()
        == """\
x = [
    # Here is a comment
    1,
    2,
]
"""
    )


def test_append_dict_to_array() -> None:
    doc = parse("x = []")
    doc["x"].append({"name": "John Doe", "email": "john@doe.com"})
    expected = 'x = [{name = "John Doe",email = "john@doe.com"}]'
    assert doc.as_string() == expected
    # Make sure the produced string is valid
    assert parse(doc.as_string()) == doc


def test_dicts_are_converted_to_tables() -> None:
    t = item({"foo": {"bar": "baz"}})

    assert (
        t.as_string()
        == """[foo]
bar = "baz"
"""
    )


def test_array_add_line() -> None:
    t = api.array()
    t.add_line(1, 2, 3, comment="Line 1")
    t.add_line(4, 5, 6, comment="Line 2")
    t.add_line(7, api.ws(","), api.ws(" "), 8, add_comma=False)
    t.add_line(comment="Line 4")
    t.add_line(indent="")
    assert len(t) == 8
    assert list(t) == [1, 2, 3, 4, 5, 6, 7, 8]
    assert (
        t.as_string()
        == """[
    1, 2, 3, # Line 1
    4, 5, 6, # Line 2
    7, 8,
    # Line 4
]"""
    )


def test_array_add_line_multiline_comment_is_rejected() -> None:
    t = api.array()
    with pytest.raises(ValueError, match="line breaks"):
        t.add_line(1, 2, 3, comment="first line\nsecond line")
    assert t.as_string() == "[]"


def test_array_add_line_invalid_value() -> None:
    t = api.array()
    with pytest.raises(ValueError, match="is not allowed"):
        t.add_line(1, api.ws(" "))
    with pytest.raises(ValueError, match="is not allowed"):
        t.add_line(Comment(Trivia("  ", comment="test")))
    assert len(t) == 0


def test_dicts_are_converted_to_tables_and_keep_order() -> None:
    t = item(
        {
            "foo": {
                "bar": "baz",
                "abc": 123,
                "baz": [{"c": 3, "b": 2, "a": 1}],
            },
        }
    )

    assert (
        t.as_string()
        == """[foo]
bar = "baz"
abc = 123

[[foo.baz]]
c = 3
b = 2
a = 1
"""
    )


def test_dicts_are_converted_to_tables_and_are_sorted_if_requested() -> None:
    t = item(
        {
            "foo": {
                "bar": "baz",
                "abc": 123,
                "baz": [{"c": 3, "b": 2, "a": 1}],
            },
        },
        _sort_keys=True,
    )

    assert (
        t.as_string()
        == """[foo]
abc = 123
bar = "baz"

[[foo.baz]]
a = 1
b = 2
c = 3
"""
    )


def test_dicts_with_sub_dicts_are_properly_converted() -> None:
    t = item(
        {"foo": {"bar": {"string": "baz"}, "int": 34, "float": 3.14}}, _sort_keys=True
    )

    assert (
        t.as_string()
        == """[foo]
float = 3.14
int = 34

[foo.bar]
string = "baz"
"""
    )


def test_item_array_of_dicts_converted_to_aot() -> None:
    a = item({"foo": [{"bar": "baz"}]})

    assert (
        a.as_string()
        == """[[foo]]
bar = "baz"
"""
    )


def test_add_float_to_int() -> None:
    content = "[table]\nmy_int = 2043"
    doc = parse(content)
    doc["table"]["my_int"] += 5.0
    assert doc["table"]["my_int"] == 2048.0
    assert isinstance(doc["table"]["my_int"], float)


def test_sub_float_from_int() -> None:
    content = "[table]\nmy_int = 2048"
    doc = parse(content)
    doc["table"]["my_int"] -= 5.0
    assert doc["table"]["my_int"] == 2043.0
    assert isinstance(doc["table"]["my_int"], float)


def test_sub_int_from_float() -> None:
    content = "[table]\nmy_int = 2048.0"
    doc = parse(content)
    doc["table"]["my_int"] -= 5
    assert doc["table"]["my_int"] == 2043.0


def test_add_sum_int_with_float() -> None:
    content = "[table]\nmy_int = 2048.3"
    doc = parse(content)
    doc["table"]["my_int"] += 5
    assert doc["table"]["my_int"] == 2053.3


def test_integers_behave_like_ints() -> None:
    i = item(34)

    assert i == 34
    assert i.as_string() == "34"

    i += 1
    assert i == 35
    assert i.as_string() == "35"

    i -= 2
    assert i == 33
    assert i.as_string() == "33"

    f = i / 2
    assert f == 16.5
    assert f.as_string() == "16.5"

    doc = parse("int = +34")
    doc["int"] += 1

    assert doc.as_string() == "int = +35"


def test_floats_behave_like_floats() -> None:
    i = item(34.12)

    assert i == 34.12
    assert i.as_string() == "34.12"

    i += 1
    assert i == 35.12
    assert i.as_string() == "35.12"

    i -= 2
    assert i == 33.12
    assert i.as_string() == "33.12"

    doc = parse("float = +34.12")
    doc["float"] += 1

    assert doc.as_string() == "float = +35.12"


def test_datetimes_behave_like_datetimes(tz_utc: tzinfo, tz_pst: tzinfo) -> None:
    i = item(datetime(2018, 7, 22, 12, 34, 56))

    assert i == datetime(2018, 7, 22, 12, 34, 56)
    assert i.as_string() == "2018-07-22T12:34:56"

    i += timedelta(days=1)
    assert i == datetime(2018, 7, 23, 12, 34, 56)
    assert i.as_string() == "2018-07-23T12:34:56"

    i -= timedelta(days=2)
    assert i == datetime(2018, 7, 21, 12, 34, 56)
    assert i.as_string() == "2018-07-21T12:34:56"

    i = i.replace(year=2019, tzinfo=tz_utc)
    assert i == datetime(2019, 7, 21, 12, 34, 56, tzinfo=tz_utc)
    assert i.as_string() == "2019-07-21T12:34:56+00:00"

    i = i.astimezone(tz_pst)
    assert i == datetime(2019, 7, 21, 4, 34, 56, tzinfo=tz_pst)
    assert i.as_string() == "2019-07-21T04:34:56-08:00"

    doc = parse("dt = 2018-07-22T12:34:56-05:00")
    doc["dt"] += timedelta(days=1)

    assert doc.as_string() == "dt = 2018-07-23T12:34:56-05:00"


def test_dates_behave_like_dates() -> None:
    i = item(date(2018, 7, 22))

    assert i == date(2018, 7, 22)
    assert i.as_string() == "2018-07-22"

    i += timedelta(days=1)
    assert i == date(2018, 7, 23)
    assert i.as_string() == "2018-07-23"

    i -= timedelta(days=2)
    assert i == date(2018, 7, 21)
    assert i.as_string() == "2018-07-21"

    i = i.replace(year=2019)
    assert i == date(2019, 7, 21)
    assert i.as_string() == "2019-07-21"

    doc = parse("dt = 2018-07-22 # Comment")
    doc["dt"] += timedelta(days=1)

    assert doc.as_string() == "dt = 2018-07-23 # Comment"


def test_parse_datetime_followed_by_space() -> None:
    # issue #260
    doc = parse("dt = 2018-07-22 ")
    assert doc["dt"] == date(2018, 7, 22)
    assert doc.as_string() == "dt = 2018-07-22 "

    doc = parse("dt = 2013-01-24 13:48:01.123456 ")
    assert doc["dt"] == datetime(2013, 1, 24, 13, 48, 1, 123456)
    assert doc.as_string() == "dt = 2013-01-24 13:48:01.123456 "


def test_times_behave_like_times() -> None:
    i = item(time(12, 34, 56))

    assert i == time(12, 34, 56)
    assert i.as_string() == "12:34:56"

    i = i.replace(hour=13)
    assert i == time(13, 34, 56)
    assert i.as_string() == "13:34:56"


def test_strings_behave_like_strs() -> None:
    i = item("foo")

    assert i == "foo"
    assert i.as_string() == '"foo"'

    i += " bar"
    assert i == "foo bar"
    assert i.as_string() == '"foo bar"'

    i += " é"
    assert i == "foo bar é"
    assert i.as_string() == '"foo bar é"'

    doc = parse('str = "foo" # Comment')
    doc["str"] += " bar"

    assert doc.as_string() == 'str = "foo bar" # Comment'


def test_string_add_preserve_escapes() -> None:
    i: Any = api.value('"foo\\"bar"')
    i += " baz"

    assert i == 'foo"bar baz'
    assert i.as_string() == '"foo\\"bar baz"'


def test_tables_behave_like_dicts() -> None:
    t = item({"foo": "bar"})

    assert (
        t.as_string()
        == """foo = "bar"
"""
    )

    t.update({"bar": "baz"})

    assert (
        t.as_string()
        == """foo = "bar"
bar = "baz"
"""
    )

    t.update({"bar": "boom"})

    assert (
        t.as_string()
        == """foo = "bar"
bar = "boom"
"""
    )

    assert t.get("bar") == "boom"
    assert t.setdefault("foobar", "fuzz") == "fuzz"
    assert (
        t.as_string()
        == """foo = "bar"
bar = "boom"
foobar = "fuzz"
"""
    )


def test_items_are_pickable() -> None:
    n: Item = item(12)

    s = pickle.dumps(n)
    assert pickle.loads(s).as_string() == "12"

    n = item(12.34)

    s = pickle.dumps(n)
    assert pickle.loads(s).as_string() == "12.34"

    n = item(True)

    s = pickle.dumps(n)
    assert pickle.loads(s).as_string() == "true"

    n = item(datetime(2018, 10, 11, 12, 34, 56, 123456))

    s = pickle.dumps(n)
    assert pickle.loads(s).as_string() == "2018-10-11T12:34:56.123456"

    n = item(date(2018, 10, 11))

    s = pickle.dumps(n)
    assert pickle.loads(s).as_string() == "2018-10-11"

    n = item(time(12, 34, 56, 123456))

    s = pickle.dumps(n)
    assert pickle.loads(s).as_string() == "12:34:56.123456"

    n = item([1, 2, 3])

    s = pickle.dumps(n)
    assert pickle.loads(s).as_string() == "[1, 2, 3]"

    n = item({"foo": "bar"})

    s = pickle.dumps(n)
    assert pickle.loads(s).as_string() == 'foo = "bar"\n'

    n = api.inline_table()
    n["foo"] = "bar"

    s = pickle.dumps(n)
    assert pickle.loads(s).as_string() == '{foo = "bar"}'

    n = item("foo")

    s = pickle.dumps(n)
    assert pickle.loads(s).as_string() == '"foo"'

    n = item([{"foo": "bar"}])

    s = pickle.dumps(n)
    assert pickle.loads(s).as_string() == 'foo = "bar"\n'


def test_trim_comments_when_building_inline_table() -> None:
    table = api.inline_table()
    row = parse('foo = "bar"  # Comment')
    table.update(row)
    assert table.as_string() == '{foo = "bar"}'
    value = item("foobaz")
    value.comment("Another comment")
    table.append("baz", value)
    assert "# Another comment" not in table.as_string()
    assert table.as_string() == '{foo = "bar", baz = "foobaz"}'


def test_comment_method_multiline_comment_is_rejected() -> None:
    doc = api.document()
    doc["x"] = 1
    with pytest.raises(ValueError, match="line breaks"):
        doc["x"].comment("first line\nsecond line")
    assert doc.as_string() == "x = 1\n"


def test_comment_method_keeps_existing_hash_prefix() -> None:
    doc = api.document()
    doc["x"] = 1
    doc["x"].comment("# already a comment")
    assert doc.as_string() == "x = 1 # already a comment\n"


def test_append_table_to_inline_table_raises() -> None:
    table = api.table()
    table.append("a", 1)
    inline_table = api.inline_table()

    with pytest.raises(ValueError, match="cannot contain a table"):
        inline_table.append("table", table)
    with pytest.raises(ValueError, match="cannot contain a table"):
        inline_table["table"] = table


def test_deleting_inline_table_element_does_not_leave_trailing_separator() -> None:
    table = api.inline_table()
    table["foo"] = "bar"
    table["baz"] = "boom"

    assert table.as_string() == '{foo = "bar", baz = "boom"}'

    del table["baz"]

    assert table.as_string() == '{foo = "bar"}'

    table = api.inline_table()
    table["foo"] = "bar"

    del table["foo"]

    table["baz"] = "boom"

    assert table.as_string() == '{baz = "boom"}'


def test_deleting_inline_table_element_does_not_leave_trailing_separator2() -> None:
    doc = parse('a = {foo = "bar", baz = "boom"}')
    table = doc["a"]
    assert table.as_string() == '{foo = "bar", baz = "boom"}'

    del table["baz"]
    assert table.as_string() == '{foo = "bar" }'

    del table["foo"]
    assert table.as_string() == "{ }"

    table["baz"] = "boom"

    assert table.as_string() == '{ baz = "boom"}'


def test_appending_to_parsed_inline_table_preserves_separator() -> None:
    doc = parse("a = { foo = 1, bar = 2 }\n")
    doc["a"]["baz"] = 3

    assert doc.as_string() == "a = { foo = 1, bar = 2, baz = 3}\n"
    parse(doc.as_string())


def test_append_key_after_inline_table_trailing_comment() -> None:
    doc = parse("tbl = {\n    p = { k = 1 },\n    q = { k = 2 }  # comment\n}\n")
    doc["tbl"]["added"] = 3

    # The separator for the new key must be placed after the previous value,
    # not after the trailing comment -- otherwise the comma becomes part of the
    # comment and the result no longer round-trips. See #512.
    rendered = doc.as_string()
    assert "# comment," not in rendered
    assert parse(rendered) == {"tbl": {"p": {"k": 1}, "q": {"k": 2}, "added": 3}}
    assert parse(rendered).as_string() == rendered


def test_deleting_inline_table_middle_element_does_not_leave_double_separator() -> None:
    doc = parse("a = {foo = 1, bar = 2, baz = 3}\n")
    del doc["a"]["bar"]

    # The dangling separator left by the removed key must not produce an
    # invalid ``, ,`` sequence; the result must round-trip.
    rendered = doc.as_string()
    assert ", ," not in rendered
    assert ",," not in rendered
    assert parse(rendered) == {"a": {"foo": 1, "baz": 3}}
    assert parse(rendered).as_string() == rendered


def test_inline_table_render_after_edits() -> None:
    # InlineTable.as_string() precomputes the last-key / last-Null indices in a
    # single pass instead of rescanning the tail on every separator comma.
    # Deleting keys (which leaves Null placeholders and dangling separators) is
    # the path that exercises those lookups, so pin the exact rendered output.
    def edited(src: str, *dels: str) -> str:
        doc = parse(src)
        for key in dels:
            del doc["t"][key]
        out = doc.as_string()
        # whatever the spacing, the result must be valid and round-trip
        assert ",," not in out and ", ," not in out
        assert parse(out).as_string() == out
        return out

    assert edited("t = {a = 1, b = 2, c = 3}", "c") == "t = {a = 1, b = 2 }"
    assert edited("t = {a = 1, b = 2, c = 3}", "b") == "t = {a = 1,  c = 3}"
    assert edited("t = {a = 1, b = 2}", "b") == "t = {a = 1 }"
    assert edited("t = {a = 1, b = 2}", "a") == "t = { b = 2}"
    assert edited("t = {a = 1, b = 2, c = 3}", "b", "c") == "t = {a = 1  }"


def test_adding_to_dotted_key_inside_inline_table() -> None:
    doc = parse("a = {b.c = 1}\n")
    doc["a"]["b"]["d"] = 2

    # The added key must stay attached to the ``b.`` prefix and be separated
    # from the existing pair; the result must round-trip.
    rendered = doc.as_string()
    assert rendered == "a = {b.c = 1, b.d = 2}\n"
    assert parse(rendered) == {"a": {"b": {"c": 1, "d": 2}}}
    assert parse(rendered).as_string() == rendered


def test_adding_to_nested_dotted_key_inside_inline_table() -> None:
    doc = parse("a = {b.c.e = 1}\n")
    doc["a"]["b"]["c"]["g"] = 2

    rendered = doc.as_string()
    assert rendered == "a = {b.c.e = 1, b.c.g = 2}\n"
    assert parse(rendered) == {"a": {"b": {"c": {"e": 1, "g": 2}}}}
    assert parse(rendered).as_string() == rendered


def test_appending_to_comma_first_array_does_not_double_separator() -> None:
    doc = parse(
        """\
a = [
      1 # one
     ,2
    ]
"""
    )
    doc["a"].append(99)

    # The comma separating ``2`` from ``99`` is carried by the new item's
    # comma-first indent; adding a trailing comma to ``2`` as well would
    # produce an invalid ``2,\\n,99`` sequence.
    rendered = doc.as_string()
    assert parse(rendered)["a"] == [1, 2, 99]
    assert parse(rendered).as_string() == rendered


def test_inserting_into_comma_first_array_does_not_double_separator() -> None:
    doc = parse(
        """\
a = [
      1 # one
     ,2
    ]
"""
    )
    doc["a"].insert(1, 99)

    rendered = doc.as_string()
    assert parse(rendered)["a"] == [1, 99, 2]
    assert parse(rendered).as_string() == rendered

    # Inserting at the front copies the first item's comma-less indent and
    # must keep its usual trailing comma.
    doc["a"].insert(0, 0)
    rendered = doc.as_string()
    assert parse(rendered)["a"] == [0, 1, 99, 2]
    assert parse(rendered).as_string() == rendered


def test_booleans_comparison() -> None:
    boolean = Bool(True, Trivia())

    assert boolean

    boolean = Bool(False, Trivia())

    assert not boolean

    s = """[foo]
value = false
"""

    content = parse(s)
    assert content["foo"]["value"] is False
    assert isinstance(content["foo"].item("value"), Bool)

    assert {"foo": {"value": False}} == content
    assert {"value": False} == content["foo"]


def test_table_copy() -> None:
    table = item({"foo": "bar"})
    table_copy = table.copy()
    assert isinstance(table_copy, Table)
    table["foo"] = "baz"
    assert table_copy["foo"] == "bar"
    assert table_copy.as_string() == 'foo = "bar"\n'


def test_copy_copy() -> None:
    result = parse(
        """
    [tool.poetry]
    classifiers = [
    # comment
        "a",
        "b",
    ]
    """
    )
    classifiers = result["tool"]["poetry"]["classifiers"]
    new = copy.copy(classifiers)
    assert new == classifiers


@pytest.mark.parametrize(
    "key_str,escaped",
    [("\\", '"\\\\"'), ('"', '"\\""'), ("\t", '"\\t"'), ("\x10", '"\\u0010"')],
)
def test_escape_key(key_str: str, escaped: str) -> None:
    assert api.key(key_str).as_string() == escaped


def test_custom_encoders() -> None:
    import decimal

    @api.register_encoder
    def encode_decimal(obj: Any) -> Item:
        if isinstance(obj, decimal.Decimal):
            return api.float_(str(obj))
        raise TypeError

    assert api.item(decimal.Decimal("1.23")).as_string() == "1.23"

    with pytest.raises(TypeError):
        api.item(object())

    assert api.dumps({"foo": decimal.Decimal("1.23")}) == "foo = 1.23\n"
    api.unregister_encoder(encode_decimal)


def test_custom_encoders_with_parent_and_sort_keys() -> None:
    """Test that custom encoders can receive _parent and _sort_keys parameters."""
    import decimal

    parent_captured = None
    sort_keys_captured = None

    @api.register_encoder
    def encode_decimal_with_context(
        obj: Any, _parent: Item | None = None, _sort_keys: bool = False
    ) -> Item:
        nonlocal parent_captured, sort_keys_captured
        if isinstance(obj, decimal.Decimal):
            parent_captured = _parent
            sort_keys_captured = _sort_keys
            return api.float_(str(obj))
        raise TypeError

    # Test with default parameters
    result = api.item(decimal.Decimal("1.23"))
    assert result.as_string() == "1.23"
    assert parent_captured is None
    assert sort_keys_captured is False

    # Test with custom parent and sort_keys
    parent_captured = None
    sort_keys_captured = None
    table = api.table()
    result = item(decimal.Decimal("4.56"), _parent=table, _sort_keys=True)
    assert result.as_string() == "4.56"
    assert parent_captured is table
    assert sort_keys_captured is True

    api.unregister_encoder(encode_decimal_with_context)


def test_custom_encoders_backward_compatibility() -> None:
    """Test that old-style custom encoders still work without modification."""
    import decimal

    @api.register_encoder
    def encode_decimal_old_style(obj: Any) -> Item:
        # Old style encoder - only accepts obj parameter
        if isinstance(obj, decimal.Decimal):
            return api.float_(str(obj))
        raise TypeError

    # Should work exactly as before
    result = api.item(decimal.Decimal("2.34"))
    assert result.as_string() == "2.34"

    # Should work when called from item() with extra parameters
    table = api.table()
    result = item(decimal.Decimal("5.67"), _parent=table, _sort_keys=True)
    assert result.as_string() == "5.67"

    api.unregister_encoder(encode_decimal_old_style)


def test_custom_encoders_with_kwargs() -> None:
    """Test that custom encoders can use **kwargs to accept additional parameters."""
    import decimal

    kwargs_captured = None

    @api.register_encoder
    def encode_decimal_with_kwargs(obj: Any, **kwargs: Any) -> Item:
        nonlocal kwargs_captured
        if isinstance(obj, decimal.Decimal):
            kwargs_captured = kwargs
            return api.float_(str(obj))
        raise TypeError

    # Test with parent and sort_keys passed as kwargs
    table = api.table()
    result = item(decimal.Decimal("7.89"), _parent=table, _sort_keys=True)
    assert result.as_string() == "7.89"
    assert kwargs_captured == {"_parent": table, "_sort_keys": True}

    api.unregister_encoder(encode_decimal_with_kwargs)


def test_custom_encoders_for_complex_objects() -> None:
    """Test custom encoders that need to encode nested structures."""

    class CustomDict:
        def __init__(self, data: dict[str, Any]) -> None:
            self.data = data

    @api.register_encoder
    def encode_custom_dict(
        obj: Any, _parent: Item | None = None, _sort_keys: bool = False
    ) -> Item:
        if isinstance(obj, CustomDict):
            # Create a table and use item() to convert nested values
            table = api.table()
            for key, value in obj.data.items():
                # Pass along _parent and _sort_keys when converting nested values
                table[key] = item(value, _parent=table, _sort_keys=_sort_keys)
            return table
        raise TypeError

    # Test with nested structure
    custom_obj = CustomDict({"a": 1, "b": {"c": 2, "d": 3}})
    result = item(custom_obj, _sort_keys=True)

    # Should properly format as a table with sorted keys
    expected = """a = 1

[b]
c = 2
d = 3
"""
    assert result.as_string() == expected

    api.unregister_encoder(encode_custom_dict)


def test_no_extra_minus_sign() -> None:
    doc = parse("a = -1")
    assert doc.as_string() == "a = -1"
    doc["a"] *= -1
    assert doc.as_string() == "a = +1"
    doc["a"] *= -1
    assert doc.as_string() == "a = -1"

    doc = parse("a = -1.5")
    assert doc.as_string() == "a = -1.5"
    doc["a"] *= -1
    assert doc.as_string() == "a = +1.5"
    doc["a"] *= -1
    assert doc.as_string() == "a = -1.5"


def test_serialize_table_with_dotted_key() -> None:
    child = api.table()
    child.add(api.key(("b", "c")), 1)
    parent = api.table()
    parent.add("a", child)
    assert parent.as_string() == "[a]\nb.c = 1\n"


def test_not_showing_parent_header_for_super_table() -> None:
    doc = api.document()

    def add_table(parent: Any, name: str) -> Any:
        parent.add(name, api.table())
        return parent[name]

    root = add_table(doc, "root")
    add_table(root, "first")
    add_table(root, "second")
    assert doc.as_string() == "[root.first]\n\n[root.second]\n"


def test_removal_of_arrayitem_with_extra_whitespace() -> None:
    expected = 'x = [\n    "bar",\n]'
    doc = parse('x = [\n    "foo" ,#spam\n    "bar",\n]')
    x = doc["x"]
    assert isinstance(x, Array)
    x.remove("foo")
    docstr = doc.as_string()
    parse(docstr)
    assert docstr == expected


def test_badly_formatted_array_and_item_removal() -> None:
    expected = """
x = [
    '0'#a
,#b

    # comment
'1' #c
    , #d
# another comment
'2'
 # yet another comment
,'3', # f
 # comments here
'4'#g
#    comments there
 ,'5' #h
    #   comments everywhere
 # so many comments
,'6' ,
      # you get a comment
# you get a comment
   '7' ,#j
# everybody gets a comment!!!
 "8"#c
,"9",  \n  "10"
]
"""
    assert expected == parse(expected).as_string()
    for i in range(11):
        doc = parse(expected)
        x = doc["x"]
        assert isinstance(x, Array)
        x.remove(str(i))
        parse(doc.as_string())


def test_array_item_removal_newline_restore_next() -> None:
    expected = "x = [\n  '0',\n    '2'\n]"

    docstr = "x = [\n  '0',\n    '1','2'\n]"
    doc = parse(docstr)
    doc["x"].remove("1")
    assert doc.as_string() == expected
    parse(doc.as_string())

    docstr = "x = [\n  '0',\n    '1',  '2'\n]"
    doc = parse(docstr)
    doc["x"].remove("1")
    assert doc.as_string() == expected
    parse(doc.as_string())


def test_table_membership_matches_inner_container() -> None:
    doc = parse('[server]\nhost = "localhost"\nport = 8080\nenabled = true\n')
    table = doc["server"]
    assert isinstance(table, Table)

    assert "host" in table
    assert "port" in table
    assert "enabled" in table
    assert "missing" not in table

    # ``str`` and ``Key`` arguments behave identically.
    assert Key("host") in table
    assert Key("missing") not in table

    # The native ``__contains__`` must agree with the inner container it delegates to.
    for key in ("host", "port", "enabled", "missing", ""):
        assert (key in table) == (key in table.value)

    # A non-string / non-Key key still raises, as before.
    with pytest.raises(TypeError):
        table.__contains__(42)


def test_inline_table_membership() -> None:
    doc = parse("point = {x = 1, y = 2}")
    inline = doc["point"]
    assert isinstance(inline, InlineTable)

    assert "x" in inline
    assert "y" in inline
    assert "z" not in inline
    assert Key("y") in inline


def test_out_of_order_table_membership() -> None:
    content = '[a.a]\nkey = "value"\n\n[a.b]\n\n[a.a.c]\n'
    doc = parse(content)
    table = doc["a"]

    # ``a`` is stored out of order (a tuple index in the inner container): the
    # only non-trivial ``__contains__`` branch, where an ``OutOfOrderTableProxy``
    # is built. Membership must be correct and must not mutate the document.
    assert "a" in table
    assert "b" in table
    assert "missing" not in table
    assert doc.as_string() == content


def test_out_of_order_table_proxy_membership() -> None:
    # A top-level table split by another table (``a`` interrupted by ``foo``)
    # resolves to an ``OutOfOrderTableProxy``; exercise its native
    # ``__contains__`` directly (the test above goes through ``Table``).
    content = "[a.x]\np = 1\n[foo]\nbar = 2\n[a.y]\nq = 3\n"
    doc = parse(content)
    table = doc["a"]
    assert isinstance(table, OutOfOrderTableProxy)

    assert "x" in table
    assert "y" in table
    assert "missing" not in table
    # a Key is accepted just like __getitem__ does
    assert Key("x") in table
    # a non-str/non-Key argument is rejected like the other mapping types
    with pytest.raises(TypeError):
        _ = 123 in table
    # membership must not resolve values or mutate the document
    assert doc.as_string() == content



import sys

import pytest

from tomlkit.exceptions import EmptyTableNameError
from tomlkit.exceptions import InternalParserError
from tomlkit.exceptions import InvalidNumberError
from tomlkit.exceptions import InvalidUnicodeValueError
from tomlkit.exceptions import ParseError
from tomlkit.exceptions import UnexpectedCharError
from tomlkit.items import Integer
from tomlkit.items import StringType
from tomlkit.parser import Parser


def test_parser_should_raise_an_internal_error_if_parsing_wrong_type_of_string() -> (
    None
):
    parser = Parser('"foo"')

    with pytest.raises(InternalParserError) as e:
        parser._parse_string(StringType.SLL)

    assert e.value.line == 1
    assert e.value.col == 0


def test_parser_should_raise_an_error_for_empty_tables() -> None:
    content = """
[one]
[]
"""

    parser = Parser(content)

    with pytest.raises(EmptyTableNameError) as e:
        parser.parse()

    assert e.value.line == 3
    assert e.value.col == 1


def test_parser_should_raise_an_error_if_equal_not_found() -> None:
    content = """[foo]
a {c = 1, d = 2}
"""
    parser = Parser(content)
    with pytest.raises(UnexpectedCharError):
        parser.parse()


def test_array_close_after_value_round_trips() -> None:
    # The parser skips the value attempt when it is sitting on the closing
    # "]" (a value-less position: an empty or trailing-comma array). This is a
    # pure optimisation -- valid arrays going through that branch must still
    # round-trip byte-for-byte.
    for content in (
        "a = []",
        "a = [ ]",
        "a = [1]",
        "a = [1,]",
        "a = [1, 2]",
        "a = [1, 2, ]",
        "a = [\n  1,\n  2,\n]",
        "a = [[1], [2]]",
        "a = [{b = 1}]",
    ):
        assert Parser(content).parse().as_string() == content


def test_malformed_array_still_raises() -> None:
    # ...and malformed arrays must still raise: the skip applies only to "]",
    # never to a real value start (which must still be parsed) nor to "," (so a
    # leading/double comma keeps erroring).
    for content in ("a = [, 1]", "a = [1 2]", "a = [1, , 2]", "a = [@]"):
        with pytest.raises(UnexpectedCharError):
            Parser(content).parse()


def test_array_with_malformed_element_raises() -> None:
    # An element that starts to parse like a value but then hits an
    # unexpected char (e.g. an inline table missing its "=" or "}") must
    # surface the error instead of being swallowed and silently dropped,
    # which left the array truncated to whatever parsed before it.
    for content in (
        "a = [{1]",
        "a = [{a]",
        "a = [1, {2]",
        "a = [{a = 1}, {2]",
        "a = [[1, {x]]",
        "a = [{a.b]",
    ):
        with pytest.raises(UnexpectedCharError):
            Parser(content).parse()


def test_parse_multiline_string_ignore_the_first_newline() -> None:
    content = 'a = """\nfoo\n"""'
    parser = Parser(content)
    assert parser.parse() == {"a": "foo\n"}

    content = 'a = """\r\nfoo\n"""'
    parser = Parser(content)
    assert parser.parse() == {"a": "foo\n"}


def test_parse_multiline_basic_string_with_crlf() -> None:
    content = 'a = """foo\r\nbar"""'
    parser = Parser(content)
    assert parser.parse() == {"a": "foo\r\nbar"}


def test_parse_multiline_literal_string_with_crlf() -> None:
    content = "a = '''foo\r\nbar'''"
    parser = Parser(content)
    assert parser.parse() == {"a": "foo\r\nbar"}


def test_multiline_string_body_round_trips() -> None:
    # The multiline body is consumed by a bulk scan that stops at the
    # delimiter / escape / CR / invalid control char, leaving raw LF and tab
    # inside the run. Bodies spanning many lines must round-trip byte-for-byte,
    # incl. mixed LF/CRLF, tabs and a lone (non-closing) quote inside.
    for delim in ('"""', "'''"):
        for body in (
            "line1\nline2\nline3\n",
            "with\ttabs\tand\nnewlines\n",
            "crlf\r\nand\r\nlf\nmixed",
            f"a lone {delim[0]} quote inside",
            f"two {delim[0] * 2} quotes inside",
            "trailing spaces   \n   leading",
        ):
            content = f"a = {delim}{body}{delim}"
            assert Parser(content).parse().as_string() == content


def test_multiline_string_still_rejects_control_chars() -> None:
    # The bulk scan must stop at a bare CR and at other control chars so the
    # per-char branches still reject them (raw LF and tab stay allowed).
    for bad in ("\r", "\x00", "\x0c", "\x1f", "\x7f"):
        content = f'a = """ok{bad}bad"""'
        with pytest.raises(ParseError):
            Parser(content).parse()


@pytest.mark.parametrize(
    "content",
    [
        r'a = "\uD800"',
        r'a = "\uDFFF"',
        r'a = "\U0000D800"',
        r'a = "\U0000DFFF"',
        r'a = "\U0000DC00"',
    ],
)
def test_parser_rejects_surrogate_unicode_escapes(content: str) -> None:
    parser = Parser(content)
    with pytest.raises(InvalidUnicodeValueError):
        parser.parse()


@pytest.mark.parametrize(
    "content",
    [
        r'a = "\u12_3"',
        r'a = "\u 123"',
        r'a = "\u+123"',
        r'a = "\u1_23"',
        r'a = "\U0010_FFFF"',
        r'a = "\U0000_0041"',
    ],
)
def test_parser_rejects_non_hex_unicode_escapes(content: str) -> None:
    parser = Parser(content)
    with pytest.raises(InvalidUnicodeValueError):
        parser.parse()


@pytest.mark.parametrize(
    "content",
    [
        "a\tb = 1",
        "[a\tb]",
        "x.y\tz = 1",
    ],
)
def test_parser_rejects_tab_in_bare_key(content: str) -> None:
    parser = Parser(content)
    with pytest.raises(ParseError):
        parser.parse()


@pytest.mark.parametrize(
    "content",
    [
        "[a.b]\n[a]\n[a.b]",
        "[a.b]\nx = 1\n[a]\n[a.b]\ny = 2",
        "[a.b.c]\n[a]\n[a.b.c]",
        "[a.b]\n[a.c]\n[a]\n[a.b]",
    ],
)
def test_parser_rejects_table_redefined_after_parent(content: str) -> None:
    parser = Parser(content)
    with pytest.raises(ParseError):
        parser.parse()


@pytest.mark.parametrize(
    "content",
    [
        "[[a]b",
        "[[a]",
        "[[a]x\ny = 1",
        "[[a.b]x",
    ],
)
def test_parser_rejects_aot_header_missing_second_bracket(content: str) -> None:
    parser = Parser(content)
    with pytest.raises(ParseError):
        parser.parse()


@pytest.mark.skipif(
    not hasattr(sys, "get_int_max_str_digits"),
    reason="requires the int-from-string digit limit (3.9.14+/3.10.7+/3.11+)",
)
def test_parser_rejects_overlong_decimal_integer() -> None:
    # A decimal integer with more digits than Python's int-from-string limit
    # raises ValueError in int(); it must be reported as an invalid number, not
    # silently fall through to float() and become inf.
    parser = Parser("a = " + "9" * 4301)
    with pytest.raises(InvalidNumberError):
        parser.parse()

    # the value just under the limit is still a normal integer
    value = Parser("a = " + "9" * 4300).parse()["a"]
    assert isinstance(value, Integer)



import json
import os

from typing import Any
from typing import Callable

import pytest

from tomlkit import load
from tomlkit import parse
from tomlkit._compat import decode
from tomlkit._utils import parse_rfc3339
from tomlkit.exceptions import TOMLKitError


TESTS_ROOT = os.path.join(os.path.dirname(__file__), "toml-test", "tests")
FILES_LIST = os.path.join(TESTS_ROOT, "files-toml-1.1.0")


def to_bool(s: str) -> bool:
    assert s in ["true", "false"]

    return s == "true"


stypes: dict[str, Callable[[str], Any]] = {
    "string": str,
    "bool": to_bool,
    "integer": int,
    "float": float,
    "datetime": parse_rfc3339,
    "datetime-local": parse_rfc3339,
    "date-local": parse_rfc3339,
    "time-local": parse_rfc3339,
}


def untag(value: Any) -> Any:
    if isinstance(value, list):
        return [untag(i) for i in value]
    elif "type" in value and "value" in value and len(value) == 2:
        if value["type"] in stypes:
            val = decode(value["value"])

            return stypes[value["type"]](val)
        elif value["type"] == "array":
            return [untag(i) for i in value["value"]]
        else:
            raise Exception(f"Unsupported type {value['type']}")
    else:
        return {k: untag(v) for k, v in value.items()}


def _load_case_list() -> list[str]:
    with open(FILES_LIST, encoding="utf-8") as f:
        return [line.strip() for line in f if line.strip()]


def _build_cases() -> tuple[
    list[dict[str, str]],
    list[str],
    list[dict[str, str]],
    list[str],
    list[str],
    list[str],
]:
    valid_cases = []
    valid_ids = []
    invalid_decode_cases = []
    invalid_decode_ids = []
    invalid_encode_cases = []
    invalid_encode_ids = []

    for relpath in _load_case_list():
        full_path = os.path.join(TESTS_ROOT, relpath)
        if not relpath.endswith(".toml"):
            continue

        case_id = relpath.rsplit(".", 1)[0]

        if relpath.startswith("invalid/encoding/"):
            invalid_encode_cases.append(full_path)
            invalid_encode_ids.append(case_id)
        elif relpath.startswith("valid/"):
            with open(full_path, encoding="utf-8", newline="") as f:
                toml_content = f.read()

            json_path = full_path.rsplit(".", 1)[0] + ".json"
            with open(json_path, encoding="utf-8") as f:
                json_content = f.read()

            valid_cases.append({"toml": toml_content, "json": json_content})
            valid_ids.append(case_id)
        elif relpath.startswith("invalid/"):
            with open(full_path, encoding="utf-8", newline="") as f:
                toml_content = f.read()

            invalid_decode_cases.append({"toml": toml_content})
            invalid_decode_ids.append(case_id)

    return (
        valid_cases,
        valid_ids,
        invalid_decode_cases,
        invalid_decode_ids,
        invalid_encode_cases,
        invalid_encode_ids,
    )


(
    VALID_CASES,
    VALID_IDS,
    INVALID_DECODE_CASES,
    INVALID_DECODE_IDS,
    INVALID_ENCODE_CASES,
    INVALID_ENCODE_IDS,
) = _build_cases()


@pytest.mark.parametrize("toml11_valid_case", VALID_CASES, ids=VALID_IDS)
def test_valid_decode(toml11_valid_case: dict[str, str]) -> None:
    json_val = untag(json.loads(toml11_valid_case["json"]))
    toml_val = parse(toml11_valid_case["toml"])

    assert toml_val == json_val
    assert toml_val.as_string() == toml11_valid_case["toml"]


@pytest.mark.parametrize(
    "toml11_invalid_decode_case", INVALID_DECODE_CASES, ids=INVALID_DECODE_IDS
)
def test_invalid_decode(toml11_invalid_decode_case: dict[str, str]) -> None:
    with pytest.raises(TOMLKitError):
        parse(toml11_invalid_decode_case["toml"])


@pytest.mark.parametrize(
    "toml11_invalid_encode_case", INVALID_ENCODE_CASES, ids=INVALID_ENCODE_IDS
)
def test_invalid_encode(toml11_invalid_encode_case: str) -> None:
    with open(toml11_invalid_encode_case, encoding="utf-8") as f:
        with pytest.raises((TOMLKitError, UnicodeDecodeError)):
            load(f)



#!/usr/bin/env python3

import argparse, pathlib, shutil, re, subprocess, os, tempfile, glob, os.path

ROOT = pathlib.Path(__file__).parent
VALID_ROOT = ROOT / f"tests/valid/spec"
INVALID_ROOT = ROOT / f"tests/invalid/spec"
DECODER = ''

def gen_multi():
    for f in glob.glob(str(ROOT / 'tests/*/*/*.multi')):
        base = os.path.dirname(f[:-6])
        i = 1
        for line in open(f, 'rb').readlines():
            name = line.split(b'=')[0].strip().decode()
            if name == '' or name[0] == '#':
                continue
            if '/valid/' in f:
                name = f'{os.path.basename(f[:-6])}-{i:02}'
                i += 1

            line = re.sub(r'(?<=[^\\])\\x([0-9a-fA-F]{2})', lambda m: chr(int(m[1], 16)), line.decode())
            path = base + "/" + name + '.toml'
            with open(path, 'wb+') as fp:
                fp.write(line.encode())

            if '/valid/' in f:  # TODO: version?
                run_decoder('1.0.0', path, path[:-5] + '.json')

def gen_list():
    with open('tests/files-toml-1.0.0', 'w+') as fp:
        subprocess.run(['go', 'run', './cmd/toml-test', 'list', '-toml=1.0.0'], stdout=fp)
    with open('tests/files-toml-1.1.0', 'w+') as fp:
        subprocess.run(['go', 'run', './cmd/toml-test', 'list', '-toml=1.1.0'], stdout=fp)

def gen_spec(version, file):
    try:
        shutil.rmtree(f"{VALID_ROOT}-{version}")
    except FileNotFoundError:
        pass
    try:
        shutil.rmtree(f"{INVALID_ROOT}-{version}")
    except FileNotFoundError:
        pass

    lines = [l[:-1] for l in open(file, 'r').readlines()]
    header = "common"
    case_index = 0
    line_index = 0
    while line_index < len(lines):
        try:
            line_index, header = parse_header(line_index, lines)
        except ParseError:
            pass
        else:
            case_index = 0
            continue

        try:
            line_index, info, block = parse_block(line_index, lines)
        except ParseError:
            pass
        else:
            if info in ["toml", ""] and block.startswith("# INVALID"):
                write_invalid_case(version, header, case_index, block)
                case_index += 1
            elif info == "toml":
                if has_active_invalid(block):
                    write_invalid_case(version, header, case_index, block)
                else:
                    write_valid_case(version, header, case_index, block)
                case_index += 1
            continue

        line_index += 1

class ParseError(RuntimeError):
    pass

def parse_header(line_index, lines):
    try:
        header = lines[line_index]
        if not header:
            raise ParseError()

        line_index += 1
        dashes = lines[line_index]
        if not re.fullmatch("-+", dashes):
            raise ParseError()

        line_index += 1
        blank = lines[line_index]
        if blank:
            raise ParseError()

        line_index += 1
    except IndexError:
        raise ParseError()

    header = header.lower().replace(" ", "-").replace("/", "-")
    return line_index, header

def parse_block(line_index, lines):
    info = ""
    try:
        fence = lines[line_index]
        if not fence.startswith('```'):
            raise ParseError()
        info = fence.removeprefix('```')

        block = []
        line = ""
        while line != '```':
            block.append(line)
            line_index += 1
            line = lines[line_index]

        line_index += 1
    except IndexError:
        raise ParseError()

    return line_index, info, "\n".join(block)

def write_invalid_case(version, header, index, block):
    path = pathlib.Path(f"{INVALID_ROOT}-{version}/{header}-{index}.toml")
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(block.strip() + '\n')

def run_decoder(version, path_toml, path_json):
    env = os.environ
    if version == '1.1.0':
        env['BURNTSUSHI_TOML_110'] = '1'
    subprocess.run([DECODER], stdin=open(path_toml), stdout=open(path_json, mode='w'))
    subprocess.run(['jfmt', '-w', path_json])

def write_valid_case(version, header, index, block):
    # Strip out datetime subseconds more than ms, since that's optional
    # behaviour.
    block = re.sub(r'(:\d\d)\.9999+', r'\1.999', block)

    path = pathlib.Path(f"{VALID_ROOT}-{version}/{header}-{index}.toml")
    path_json = pathlib.Path(f"{VALID_ROOT}-{version}/{header}-{index}.json")

    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(block.strip() + '\n')

    run_decoder(version, path, path_json)

    invalid_index = 0
    lines = block.splitlines()
    for i, line in enumerate(lines):
        if "# INVALID" in line:
            new_lines = lines[:]
            assert line.startswith("# "), f"{line}"
            new_lines[i] = line.removeprefix("# ")
            write_invalid_case(version, header, f"{index}-{invalid_index}", "\n".join(new_lines))
            invalid_index += 1

def has_active_invalid(block):
    lines = block.splitlines()
    for line in lines:
        if "# INVALID" in line and not line.startswith("# "):
            return True
    return False

if __name__ == "__main__":
    with tempfile.TemporaryDirectory() as tmp:
        DECODER = os.path.join(tmp, 'toml-test-decoder')
        subprocess.run(['go', 'build', '-o', DECODER, 'github.com/BurntSushi/toml/cmd/toml-test-decoder'])

        gen_multi()
        gen_spec('1.0.0', 'specs/v1.0.0.md')
        gen_spec('1.1.0', 'specs/v1.1.0.md')
        gen_list()



from typing import Any

from tomlkit import dumps
from tomlkit import loads


def test_write_backslash() -> None:
    d = {"foo": "\\e\u25e6\r"}

    expected = """foo = "\\\\e\u25e6\\r"
"""

    assert expected == dumps(d)
    result: Any = loads(dumps(d))["foo"]
    assert result == "\\e\u25e6\r"


def test_escape_special_characters_in_key() -> None:
    d = {"foo\nbar": "baz"}
    expected = '"foo\\nbar" = "baz"\n'
    assert expected == dumps(d)
    result: Any = loads(dumps(d))["foo\nbar"]
    assert result == "baz"


def test_write_inline_table_in_nested_arrays() -> None:
    d = {"foo": [[{"a": 1}]]}
    expected = "foo = [[{a = 1}]]\n"
    assert expected == dumps(d)
    result: Any = loads(dumps(d))["foo"]
    assert result == [[{"a": 1}]]


def test_serialize_aot_with_nested_tables() -> None:
    doc = {"a": [{"b": {"c": 1}}]}
    expected = """\
[[a]]
[a.b]
c = 1
"""
    assert dumps(doc) == expected
    assert loads(expected) == doc
